@@ -11,11 +11,11 @@ use super::opcodes::{
 };
 use super::sanitize_int::sanitize_uint;
 use super::validation_error::{first, next, rest, ErrorCode, ValidationErr};
-use crate::bytes::Bytes32;
 use crate::gen::flags::COND_ARGS_NIL;
 use crate::gen::flags::NO_UNKNOWN_CONDS;
 use crate::gen::flags::STRICT_ARGS_COUNT;
 use crate::gen::validation_error::check_nil;
+use chia_protocol::bytes::Bytes32;
 use clvmr::allocator::{Allocator, NodePtr, SExp};
 use clvmr::cost::Cost;
 use clvmr::op_utils::u64_from_bytes;
@@ -305,6 +305,12 @@ impl PartialEq for NewCoin {
     }
 }
 
+// spend flags
+
+// a spend is eligible for deduplication if it does not have any AGG_SIG_ME
+// nor AGG_SIG_UNSAFE
+pub const ELIGIBLE_FOR_DEDUP: u32 = 1;
+
 // These are all the conditions related directly to a specific spend.
 #[derive(Debug)]
 pub struct Spend {
@@ -322,6 +328,8 @@ pub struct Spend {
     pub create_coin: HashSet<NewCoin>,
     // Agg Sig Me conditions
     pub agg_sig_me: Vec<(NodePtr, NodePtr)>,
+    // Flags describing properties of this spend. See flags above
+    pub flags: u32,
 }
 
 // these are all the conditions and properties of a complete spend bundle.
@@ -407,6 +415,8 @@ fn parse_spend_conditions(
         seconds_relative: 0,
         create_coin: HashSet::new(),
         agg_sig_me: Vec::new(),
+        // assume it's eligible until we see an agg-sig condition
+        flags: ELIGIBLE_FOR_DEDUP,
     };
 
     let mut iter = first(a, cond)?;
@@ -511,9 +521,11 @@ fn parse_spend_conditions(
             }
             Condition::AggSigMe(pk, msg) => {
                 spend.agg_sig_me.push((pk, msg));
+                spend.flags &= !ELIGIBLE_FOR_DEDUP;
             }
             Condition::AggSigUnsafe(pk, msg) => {
                 ret.agg_sig_unsafe.push((pk, msg));
+                spend.flags &= !ELIGIBLE_FOR_DEDUP;
             }
             Condition::Skip => {}
         }
@@ -900,6 +912,7 @@ fn test_invalid_condition_args_terminator() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(spend.seconds_relative, 50);
 }
@@ -926,6 +939,7 @@ fn test_invalid_condition_list_terminator() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(spend.seconds_relative, 50);
 }
@@ -984,6 +998,7 @@ fn test_single_seconds_relative() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(spend.seconds_relative, 101);
 }
@@ -999,6 +1014,7 @@ fn test_single_seconds_relative_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(spend.seconds_relative, 101);
 }
@@ -1037,6 +1053,7 @@ fn test_multiple_seconds_relative() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     // we use the MAX value
     assert_eq!(spend.seconds_relative, 503);
@@ -1052,6 +1069,7 @@ fn test_single_seconds_absolute() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(conds.seconds_absolute, 104);
 }
@@ -1067,6 +1085,7 @@ fn test_single_seconds_absolute_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(conds.seconds_absolute, 104);
 }
@@ -1108,6 +1127,7 @@ fn test_multiple_seconds_absolute() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     // we use the MAX value
     assert_eq!(conds.seconds_absolute, 503);
@@ -1123,6 +1143,7 @@ fn test_single_height_relative() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(spend.height_relative, Some(101));
 }
@@ -1138,6 +1159,7 @@ fn test_single_height_relative_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(spend.height_relative, Some(101));
 }
@@ -1167,6 +1189,7 @@ fn test_single_height_relative_zero() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(spend.height_relative, Some(0));
 }
@@ -1193,6 +1216,7 @@ fn test_multiple_height_relative() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     // we use the MAX value
     assert_eq!(spend.height_relative, Some(503));
@@ -1208,6 +1232,7 @@ fn test_single_height_absolute() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(conds.height_absolute, 100);
 }
@@ -1223,6 +1248,7 @@ fn test_single_height_absolute_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(conds.height_absolute, 100);
 }
@@ -1264,6 +1290,7 @@ fn test_multiple_height_absolute() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     // we use the MAX value
     assert_eq!(conds.height_absolute, 503);
@@ -1279,6 +1306,7 @@ fn test_single_reserve_fee() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(conds.reserve_fee, 100);
 }
@@ -1294,6 +1322,7 @@ fn test_single_reserve_fee_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     assert_eq!(conds.reserve_fee, 100);
 }
@@ -1337,6 +1366,7 @@ fn test_multiple_reserve_fee() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 
     // reserve fee conditions are accumulated 100 + 50 = 150
     assert_eq!(conds.reserve_fee, 175);
@@ -1355,6 +1385,7 @@ fn test_coin_announces_consume() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1373,6 +1404,7 @@ fn test_create_coin_announce_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1407,6 +1439,7 @@ fn test_assert_coin_announce_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1487,6 +1520,7 @@ fn test_puzzle_announces_consume() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1505,6 +1539,7 @@ fn test_create_puzzle_announces_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1539,6 +1574,7 @@ fn test_assert_puzzle_announces_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1618,6 +1654,7 @@ fn test_single_assert_my_amount() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1631,6 +1668,7 @@ fn test_single_assert_my_amount_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1693,6 +1731,7 @@ fn test_multiple_assert_my_amount() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1727,6 +1766,7 @@ fn test_single_assert_my_coin_id() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1740,6 +1780,7 @@ fn test_single_assert_my_coin_id_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1780,6 +1821,7 @@ fn test_multiple_assert_my_coin_id() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1816,6 +1858,7 @@ fn test_single_assert_my_parent_coin_id() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1829,6 +1872,7 @@ fn test_single_assert_my_parent_coin_id_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1856,6 +1900,7 @@ fn test_multiple_assert_my_parent_coin_id() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1891,6 +1936,7 @@ fn test_single_assert_my_puzzle_hash() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1904,6 +1950,7 @@ fn test_single_assert_my_puzzle_hash_extra_arg() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1931,6 +1978,7 @@ fn test_multiple_assert_my_puzzle_hash() {
     let spend = &conds.spends[0];
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1972,6 +2020,7 @@ fn test_single_create_coin() {
         assert_eq!(c.amount, 42_u64);
         assert_eq!(c.hint, a.null());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -1991,6 +2040,7 @@ fn test_create_coin_max_amount() {
         assert_eq!(c.amount, 0xffffffffffffffff_u64);
         assert_eq!(c.hint, a.null());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2042,6 +2092,7 @@ fn test_create_coin_with_hint() {
         assert!(c.amount == 42_u64);
         assert!(a.atom(c.hint) == H1.to_vec());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2062,6 +2113,7 @@ fn test_create_coin_extra_arg() {
         assert!(c.amount == 42_u64);
         assert!(a.atom(c.hint) == H1.to_vec());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2095,6 +2147,7 @@ fn test_create_coin_with_multiple_hints() {
         assert!(c.amount == 42_u64);
         assert!(a.atom(c.hint) == H1.to_vec());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2114,6 +2167,7 @@ fn test_create_coin_with_hint_as_atom() {
         assert_eq!(c.amount, 42_u64);
         assert_eq!(c.hint, a.null());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2132,6 +2186,7 @@ fn test_create_coin_with_invalid_hint_as_terminator() {
         assert_eq!(c.amount, 42_u64);
         assert_eq!(c.hint, a.null());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2163,6 +2218,7 @@ fn test_create_coin_with_short_hint() {
         assert!(c.amount == 42_u64);
         assert!(a.atom(c.hint) == MSG1.to_vec());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2182,6 +2238,7 @@ fn test_create_coin_with_long_hint() {
         assert_eq!(c.amount, 42_u64);
         assert_eq!(c.hint, a.null());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2202,6 +2259,7 @@ fn test_create_coin_with_pair_hint() {
         assert_eq!(c.amount, 42_u64);
         assert_eq!(a.atom(c.hint), H1.to_vec());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2221,6 +2279,7 @@ fn test_create_coin_with_cons_hint() {
         assert_eq!(c.amount, 42_u64);
         assert_eq!(c.hint, a.null());
     }
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2246,6 +2305,7 @@ fn test_multiple_create_coin() {
         amount: 43_u64,
         hint: a.null()
     }));
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2308,6 +2368,7 @@ fn test_single_agg_sig_me() {
         assert_eq!(a.atom(c.0), PUBKEY);
         assert_eq!(a.atom(c.1), MSG1);
     }
+    assert_eq!(spend.flags, 0);
 }
 
 #[test]
@@ -2329,6 +2390,7 @@ fn test_duplicate_agg_sig_me() {
         assert_eq!(a.atom(c.0), PUBKEY);
         assert_eq!(a.atom(c.1), MSG1);
     }
+    assert_eq!(spend.flags, 0);
 }
 
 #[test]
@@ -2402,6 +2464,7 @@ fn test_single_agg_sig_unsafe() {
         assert_eq!(a.atom(*pk), PUBKEY);
         assert_eq!(a.atom(*msg), MSG1);
     }
+    assert_eq!(spend.flags, 0);
 }
 
 #[test]
@@ -2469,6 +2532,7 @@ fn test_agg_sig_unsafe_invalid_terminator() {
         assert_eq!(a.atom(*pk), PUBKEY);
         assert_eq!(a.atom(*msg), MSG1);
     }
+    assert_eq!(spend.flags, 0);
 }
 
 #[test]
@@ -2503,6 +2567,7 @@ fn test_agg_sig_me_invalid_terminator() {
         assert_eq!(a.atom(*pk), PUBKEY);
         assert_eq!(a.atom(*msg), MSG1);
     }
+    assert_eq!(spend.flags, 0);
 }
 
 #[test]
@@ -2539,6 +2604,7 @@ fn test_duplicate_agg_sig_unsafe() {
         assert_eq!(a.atom(*pk), PUBKEY);
         assert_eq!(a.atom(*msg), MSG1);
     }
+    assert_eq!(spend.flags, 0);
 }
 
 #[test]
@@ -2664,6 +2730,7 @@ fn test_always_true() {
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
     assert_eq!(spend.agg_sig_me.len(), 0);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
 
 #[test]
@@ -2685,4 +2752,5 @@ fn test_always_true_with_arg() {
     assert_eq!(*spend.coin_id, test_coin_id(H1, H2, 123));
     assert_eq!(a.atom(spend.puzzle_hash), H2);
     assert_eq!(spend.agg_sig_me.len(), 0);
+    assert_eq!(spend.flags, ELIGIBLE_FOR_DEDUP);
 }
